#!/usr/bin/env python

"""
CPAchecker is a tool for configurable software verification.
This file is part of CPAchecker.

Copyright (C) 2007-2012  Dirk Beyer
All rights reserved.

Licensed under the Apache License, Version 2.0 (the "License");
you may not use this file except in compliance with the License.
You may obtain a copy of the License at

    http://www.apache.org/licenses/LICENSE-2.0

Unless required by applicable law or agreed to in writing, software
distributed under the License is distributed on an "AS IS" BASIS,
WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
See the License for the specific language governing permissions and
limitations under the License.


CPAchecker web page:
  http://cpachecker.sosy-lab.org
"""

# prepare for Python 3
from __future__ import absolute_import, print_function, unicode_literals

import sys
sys.dont_write_bytecode = True # prevent creation of .pyc files

import xml.etree.ElementTree as ET
import collections
import os.path
import glob
import json
import argparse
import subprocess
import time
import tempita
import datetime

from decimal import *

import benchmark.result as result

NAME_START = "results" # first part of filename of table

DEFAULT_OUTPUT_PATH = "test/results/"

LIB_URL = "http://www.sosy-lab.org/lib"
LIB_URL_OFFLINE = "lib/javascript"

TEMPLATE_FILE_NAME = os.path.join(os.path.dirname(__file__), 'table-generator-template.{format}')
TEMPLATE_FORMATS = ['html', 'csv']
TEMPLATE_ENCODING = 'UTF-8'


class Util:
    """
    This Class contains some useful functions for Strings, Files and Lists.
    """

    @staticmethod
    def getFileList(shortFile):
        """
        The function getFileList expands a short filename to a sorted list
        of filenames. The short filename can contain variables and wildcards.
        """

        # expand tilde and variables
        expandedFile = os.path.expandvars(os.path.expanduser(shortFile))

        # expand wildcards
        fileList = glob.glob(expandedFile)

        # sort alphabetical,
        # if list is emtpy, sorting returns None, so better do not sort
        if len(fileList) != 0:
            fileList.sort()
        else:
            print ('\nWarning: no file matches "{0}".'.format(shortFile))

        return fileList


    @staticmethod
    def extendFileList(filelist):
        '''
        This function takes a list of files, expands wildcards
        and returns a new list of files.
        '''
        return [file for wildcardFile in filelist for file in Util.getFileList(wildcardFile)]


    @staticmethod
    def containsAny(text, list):
        """
        This function returns True, iff any string in list is a substring of text.
        """
        for elem in list:
            if elem in text:
                return True
        return False


    @staticmethod
    def splitNumberAndUnit(s):
        """
        Split a string into two parts: a number prefix and an arbitrary suffix.
        Splitting is done from the end, so the split is where the last digit
        in the string is (that means the prefix may include non-digit characters,
        if they are followed by at least one digit).
        """
        if not s:
            return (s, '')
        pos = len(s)
        while pos and not s[pos-1].isdigit():
            pos -= 1
        return (s[:pos], s[pos:])


    @staticmethod
    def formatNumber(value, numberOfDigits):
        """
        If the value is a number (or number plus one char),
        this function returns a string-representation of the number
        with a number of digits after the decimal separator.
        If the number has more digits, it is rounded, else zeros are added.

        If the value is no number, it is returned unchanged.
        """
        # if the number ends with "s" or another unit, remove it
        value, suffix = Util.splitNumberAndUnit((value or '').strip())
        try:
            floatValue = float(value)
            value = "{value:.{width}f}".format(width=numberOfDigits, value=floatValue)
        except ValueError: # if value is no float, don't format it
            pass
        return value + suffix


    @staticmethod
    def toDecimal(s):
        # remove whitespaces and trailing units (e.g., in '1.23s')
        s, _ = Util.splitNumberAndUnit((s or '').strip())
        return Decimal(s) if s else Decimal()


    @staticmethod
    def collapseEqualValues(values, counts):
        """
        Take a tuple (values, counts), remove consecutive values and increment their count instead.
        """
        assert len(values) == len(counts)
        previousValue = values[0]
        previousCount = 0

        for value, count in zip(values, counts):
            if value != previousValue:
                yield (previousValue, previousCount)
                previousCount = 0
                previousValue = value
            previousCount += count

        yield (previousValue, previousCount)

    @staticmethod
    def getColumnValue(sourcefileTag, columnTitle, default=None):
        for column in sourcefileTag.findall('column'):
            if column.get('title') == columnTitle:
                    return column.get('value')
        return default

    @staticmethod
    def flatten(list):
        return [value for sublist in list for value in sublist]

    @staticmethod
    def json(obj):
        return tempita.html(json.dumps(obj))

    @staticmethod
    def prettylist(list):
        if not list:
            return ''

        # Filter out duplicate values while keeping order
        values = set()
        uniqueList = []
        for entry in list:
            if not entry in values:
                values.add(entry)
                uniqueList.append(entry)

        return uniqueList[0] if len(uniqueList) == 1 \
            else '[' + '; '.join(uniqueList) + ']'

def parseTableDefinitionFile(file):
    '''
    This function parses the input to get run sets and columns.
    The param 'file' is an XML file defining the result files and columns.

    If column titles are given in the XML file,
    they will be searched in the result files.
    If no title is given, all columns of the result file are taken.

    @return: a list of RunSetResult objects
    '''
    print ("reading table definition from '{0}'...".format(file))
    if not os.path.isfile(file):
        print ('File {0!r} does not exist.'.format(file))
        exit()

    def extractColumnsFromTableDefinitionFile(xmltag):
        """
        Extract all columns mentioned in the result tag of a table definition file.
        """
        return [Column(c.get("title"), c.text, c.get("numberOfDigits"))
                for c in xmltag.findall('column')]

    runSetResults = []
    tableGenFile = ET.ElementTree().parse(file)
    if 'table' != tableGenFile.tag:
        print ("ERROR:\n" \
            + "    The XML file seems to be invalid.\n" \
            + "    The rootelement of table-definition-file is not named 'table'.")
        exit()

    defaultColumnsToShow = extractColumnsFromTableDefinitionFile(tableGenFile)

    baseDir = os.path.dirname(file)

    def getResultTags(rootTag):
        tags = rootTag.findall('result')
        if not tags:
            tags = rootTag.findall('test')
            if tags:
                print("Warning: file {0} contains deprecated 'test' tags, rename them to 'result'".format(file))
        return tags

    for resultTag in getResultTags(tableGenFile):
        columnsToShow = extractColumnsFromTableDefinitionFile(resultTag) or defaultColumnsToShow
        filelist = Util.getFileList(os.path.join(baseDir, resultTag.get('filename'))) # expand wildcards
        runSetResults += [RunSetResult.createFromXML(resultsFile, parseResultsFile(resultsFile), columnsToShow) for resultsFile in filelist]

    for unionTag in tableGenFile.findall('union'):
        columnsToShow = extractColumnsFromTableDefinitionFile(unionTag) or defaultColumnsToShow
        result = RunSetResult([], collections.defaultdict(list), columnsToShow)

        for resultTag in getResultTags(unionTag):
            filelist = Util.getFileList(os.path.join(baseDir, resultTag.get('filename'))) # expand wildcards
            for resultsFile in filelist:
                result.append(resultsFile, parseResultsFile(resultsFile))

        if result.filelist:
            result.attributes['name'] = unionTag.get('title', unionTag.get('name', result.attributes['name']))
            runSetResults.append(result)

    return runSetResults


class Column:
    """
    The class Column contains title, pattern (to identify a line in logFile),
    and numberOfDigits of a column.
    It does NOT contain the value of a column.
    """
    def __init__(self, title, pattern, numOfDigits):
        self.title = title
        self.pattern = pattern
        self.numberOfDigits = numOfDigits


class RunSetResult():
    """
    The Class RunSetResult contains all the results of one execution of a run set:
    the sourcefiles tags (with sourcefiles + values), the columns to show
    and the benchmark attributes.
    """
    def __init__(self, filelist, attributes, columns):
        self.filelist = filelist
        self.attributes = attributes
        self.columns = columns

    def getSourceFileNames(self):
        return [file.get('name') for file in self.filelist if Util.getColumnValue(file, 'status')]

    def append(self, resultFile, resultElem):
        self.filelist += resultElem.findall('sourcefile')
        for attrib, values in RunSetResult._extractAttributesFromResult(resultFile, resultElem).items():
            self.attributes[attrib].extend(values)

        if not self.columns:
            self.columns = RunSetResult._extractExistingColumnsFromResult(resultFile, resultElem)

    @staticmethod
    def createFromXML(resultFile, resultElem, columns=None):
        '''
        This function extracts everything necessary for creating a RunSetResult object
        from the "result" XML tag of a benchmark result file.
        It returns a RunSetResult object.
        '''
        attributes = RunSetResult._extractAttributesFromResult(resultFile, resultElem)

        if not columns:
            columns = RunSetResult._extractExistingColumnsFromResult(resultFile, resultElem)

        return RunSetResult(resultElem.findall('sourcefile'),
                attributes, columns)

    @staticmethod
    def _extractExistingColumnsFromResult(resultFile, resultElem):
        if resultElem.find('sourcefile') is None:
            print("Empty resultfile found: " + resultFile)
            return []
        else: # show all available columns
            return [Column(c.get("title"), None, None)
                    for c in resultElem.find('sourcefile').findall('column')]

    @staticmethod
    def _extractAttributesFromResult(resultFile, resultTag):
        attributes = collections.defaultdict(list)

        # Defaults
        attributes['name'  ] = [resultTag.get('benchmarkname')]
        attributes['branch'] = [os.path.basename(resultFile).split('#')[0] if '#' in resultFile else '']

        # Update with real values
        for attrib, value in resultTag.attrib.items():
            attributes[attrib] = [value]

        # Add system information if present
        for systemTag in resultTag.findall('systeminfo'):
            cpuTag = systemTag.find('cpu')
            attributes['os'   ].append(systemTag.find('os').get('name'))
            attributes['cpu'  ].append(cpuTag.get('model'))
            attributes['cores'].append( cpuTag.get('cores'))
            attributes['freq' ].append(cpuTag.get('frequency'))
            attributes['ram'  ].append(systemTag.find('ram').get('size'))
            attributes['host' ].append(systemTag.get('hostname', 'unknown'))

        return attributes


def parseResultsFile(resultFile):
    '''
    This function parses a XML file with the results of the execution of a run set.
    It returns the "result" XML tag
    '''
    if not os.path.isfile(resultFile):
        print ('File {0!r} is not found.'.format(resultFile))
        exit()

    resultElem = ET.ElementTree().parse(resultFile)

    if resultElem.tag not in ['result', 'test']:
        print (("ERROR:\n" \
            + "XML file with benchmark results seems to be invalid.\n" \
            + "The root element of the file is not named 'result' or 'test'.\n" \
            + "If you want to run a table-definition file,\n"\
            + "you should use the option '-x' or '--xml'.").replace('\n','\n    '))
        exit()

    insertLogFileNames(resultFile, resultElem)
    return resultElem

def insertLogFileNames(resultFile, resultElem):
    parts = os.path.basename(resultFile).split("#", 1)

    # get folder of logfiles
    date = resultElem.get('date').replace(':','').replace(' ','_') # from ISO-format to filename-format
    logFolder = resultElem.get('benchmarkname') + '.' + date + '.logfiles/'
    if len(parts) > 1:
        logFolder = parts[0] + '#' + logFolder
    logFolder = os.path.join(os.path.dirname(resultFile), resultElem.get('baseDir', ''), logFolder)

    # append begin of filename
    runSetName = resultElem.get('name')
    if runSetName is not None:
        blockname = resultElem.get('block')
        if blockname is None:
            logFolder += runSetName + "."
        elif blockname == runSetName:
            pass # real runSetName is empty
        else:
            assert runSetName.endswith("." + blockname)
            runSetName = runSetName[:-(1 + len(blockname))] # remove last chars
            logFolder += runSetName + "."

    # for each file: append original filename and insert logFileName into sourcefileElement
    for sourcefile in resultElem.findall('sourcefile'):
        logFileName = os.path.basename(sourcefile.get('name')) + ".log"
        sourcefile.logfile = logFolder + logFileName

def getDefaultLogFolder(resultElem):
    return logFolder


def mergeSourceFiles(runSetResults):
    """
    This function merges the filelists of all RunSetResult objects.
    If necessary, it can merge lists of names: [A,C] + [A,B] --> [A,B,C]
    and add dummy elements to the filelists.
    It also ensures the same order of files.
    Returns a list of filenames
    """
    nameList = []
    nameSet = set()
    for result in runSetResults:
        index = -1
        currentResultNameSet = set()
        for name in result.getSourceFileNames():
            if name in currentResultNameSet:
                print ("File {0} is present twice, skipping it.".format(name))
            else:
                currentResultNameSet.add(name)
                if name not in nameSet:
                    nameList.insert(index+1, name)
                    nameSet.add(name)
                    index += 1
                else:
                    index = nameList.index(name)

    mergeFilelists(runSetResults, nameList)
    return nameList

def mergeFilelists(runSetResults, filenames):
    """
    Set the filelists of all RunSetResult elements so that they contain the same files
    in the same order. For missing files a dummy element is inserted.
    """
    for result in runSetResults:
        # create mapping from name to sourcefile tag
        dic = dict([(file.get('name'), file) for file in result.filelist])
        result.filelist = [] # clear and repopulate filelist
        for filename in filenames:
            fileResult = dic.get(filename)
            if fileResult == None:
                fileResult = ET.Element('sourcefile') # create an empty dummy element
                fileResult.logfile = None
                print ('    no result for {0}'.format(filename))
            result.filelist.append(fileResult)


def findCommonSourceFiles(runSetResults):
    filesInFirstRunSet = runSetResults[0].getSourceFileNames()

    fileSet = set(filesInFirstRunSet)
    for result in runSetResults:
        fileSet = fileSet & set(result.getSourceFileNames())

    fileList = []
    if not fileSet:
        print('No files are present in all benchmark results.')
    else:
        fileList = [file for file in filesInFirstRunSet if file in fileSet]
        mergeFilelists(runSetResults, fileList)

    return fileList


class RunResult:
    """
    The class RunResult contains the results of a single verification run.
    """
    def __init__(self, status, category, logFile, columns, values):
        assert(len(columns) == len(values))
        self.status = status
        self.logFile = logFile
        self.columns = columns
        self.values = values
        self.category = category

    @staticmethod
    def createFromXML(sourcefileTag, listOfColumns, correctOnly):
        '''
        This function collects the values from one run.
        Only columns that should be part of the table are collected.
        '''

        def readLogfileLines(logfileName):
            if not logfileName: return []
            try:
                with open(logfileName, 'rt') as logfile:
                    return logfile.readlines()
            except IOError as e:
                print('WARNING: Could not read value from logfile: {}'.format(e))
                return []

        def getValueFromLogfile(lines, identifier):
            """
            This method searches for values in lines of the content.
            The format of such a line must be:    "identifier:  value  (rest)".

            If a value is not found, the value is set to None.
            """
            # stop after the first line, that contains the searched text
            for line in lines:
                if identifier in line:
                    startPosition = line.find(':') + 1
                    endPosition = line.find('(', startPosition) # bracket maybe not found -> (-1)
                    if (endPosition == -1):
                        return line[startPosition:].strip()
                    else:
                        return line[startPosition: endPosition].strip()
            return None

        status = Util.getColumnValue(sourcefileTag, 'status', '')
        category = result.getResultCategory(sourcefileTag.get('name'), status)
        score = result.calculateScore(category)
        logfileLines = None

        if status == '':
            values = [''] * len(listOfColumns)
            return RunResult(status, category, sourcefileTag.logfile, listOfColumns, values)

        values = []

        for column in listOfColumns: # for all columns that should be shown
            value = None # default value
            if column.title.lower() == 'score':
                value = str(score)
            elif column.title.lower() == 'status':
                value = status

            elif not correctOnly or score > 0:
                if not column.pattern: # collect values from XML
                    value = Util.getColumnValue(sourcefileTag, column.title)

                else: # collect values from logfile
                    if logfileLines is None: # cache content
                        logfileLines = readLogfileLines(sourcefileTag.logfile)

                    value = getValueFromLogfile(logfileLines, column.pattern)

            if column.numberOfDigits is not None:
                value = Util.formatNumber(value, column.numberOfDigits)

            values.append(value)

        return RunResult(status, category, sourcefileTag.logfile, listOfColumns, values)


class Row:
    """
    The class Row contains all the results for one sourcefile (a list of RunResult instances).
    It corresponds to one complete row in the final tables.
    """
    def __init__(self, fileName):
        self.fileName = fileName
        self.results = []

    def addRunResult(self, runresult):
        self.results.append(runresult)

    def setRelativePath(self, commonPrefix, baseDir):
        """
        generate output representation of rows
        """
        # make path relative to directory of output file if necessary
        self.filePath = self.fileName if os.path.isabs(self.fileName) \
                                 else os.path.relpath(self.fileName, baseDir)

        self.shortFileName = self.fileName.replace(commonPrefix, '', 1)

def rowsToColumns(rows):
    """
    Convert a list of Rows into a column-wise list of list of RunResult
    """
    return zip(*[row.results for row in rows])


def getRows(runSetResults, fileNames, correctOnly):
    """
    Create list of rows with all data. Each row consists of several RunResults.
    """
    rows = [Row(fileName) for fileName in fileNames]

    # get values for each run set
    for result in runSetResults:
        # get values for each file in a run set
        for fileResult, row in zip(result.filelist, rows):
            row.addRunResult(RunResult.createFromXML(fileResult, result.columns, correctOnly))

    return rows

def getCategoryCount(categoryList):
    # count different elems in statusList
    counts = collections.defaultdict(int)
    for category in categoryList:
        counts[category] += 1

    return (counts[result.RESULT_CORRECT_TRUE],
            counts[result.RESULT_CORRECT_FALSE],
            counts[result.RESULT_CORRECT_PROP_DEREF],
            counts[result.RESULT_CORRECT_PROP_FREE],
            counts[result.RESULT_CORRECT_PROP_MEMTRACK],
            counts[result.RESULT_WRONG_TRUE],
            counts[result.RESULT_WRONG_FALSE],
            counts[result.RESULT_WRONG_PROP_DEREF],
            counts[result.RESULT_WRONG_PROP_FREE],
            counts[result.RESULT_WRONG_PROP_MEMTRACK])

def getCounts(rows): # for options.dumpCounts
    countsList = {}

    for runResults in rowsToColumns(rows):
        statusList = [runResult.category for runResult in runResults]
        resultList = [runResult.status for runResult in runResults]
        resultSet = set(resultList)

        for label in ("true", "false", "false(valid-deref)", "false(valid-free)", "false(valid-memtrack)"):
          if label in resultSet:
            resultSet.remove(label)

        corrSafe, corrUnsafe, corrDeref, corrFree, corrMemtrack, wrongSafe, wrongUnsafe, wrongDeref, wrongFree, wrongMemtrack = getCategoryCount(statusList)
        countsList["CORRECT SAFE"] = corrSafe
        countsList["CORRECT UNSAFE"] = corrUnsafe + corrDeref + corrFree + corrMemtrack
        countsList["INCORRECT_SAFE"] = wrongSafe
        countsList["INCORRECT UNSAFE"] = wrongUnsafe + wrongDeref + wrongFree + wrongMemtrack

        countsList["SCORE"] = corrSafe * 2
        countsList["SCORE"] += corrUnsafe + corrDeref + corrFree + corrMemtrack
        countsList["SCORE"] -= (wrongUnsafe + wrongDeref + wrongFree + wrongMemtrack) * 4
        countsList["SCORE"] -= (wrongSafe * 8)

        for key in resultSet:
          countsList[key] = resultList.count(key)

    return countsList

def main(args=None):

    if args is None:
        args = sys.argv

    parser = argparse.ArgumentParser(
        description="""Create table with the results of one or more benchmark executions.
        Documented example files for the table definitions can be found in 'doc/examples'\n"""
    )

    parser.add_argument("table",
        metavar="RESULT",
        type=str,
        help="XML file with the results from the benchmark script"
    )
    options = parser.parse_args(args[1:])

    if options.table:
      inputFile = options.table

    runSetResult = [RunSetResult.createFromXML(inputFile, parseResultsFile(inputFile))]
    name = os.path.basename(inputFile)

    if not runSetResult:
        print ('\nError! No file with benchmark results found.')
        exit()

    fileNames = mergeSourceFiles(runSetResult)
    rows     = getRows(runSetResult, fileNames, False)
    if not rows:
        print ('Warning: No results found, no tables produced.')
        sys.exit()

    print("DATE : %s" % datetime.datetime.today().date())
    countsList = getCounts(rows)
    for counts in countsList:
        print ("%s : %d" % (counts, countsList[counts]))


if __name__ == '__main__':
    try:
        sys.exit(main())
    except KeyboardInterrupt:
        print ('script was interrupted by user')
        pass
